package eu.jonahbauer.tichu.common.model;

import lombok.RequiredArgsConstructor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Range;
import org.jetbrains.annotations.Unmodifiable;

import java.util.*;
import java.util.stream.Collectors;

@RequiredArgsConstructor
public enum Card {
    HOUND(0, null),
    MAHJONG(1, null),
    PHOENIX(null, null),
    DRAGON(15, null),
    GREEN_2(2, Suit.GREEN),      RED_2(2, Suit.RED),      BLACK_2(2, Suit.BLACK),      BLUE_2(2, Suit.BLUE),
    GREEN_3(3, Suit.GREEN),      RED_3(3, Suit.RED),      BLACK_3(3, Suit.BLACK),      BLUE_3(3, Suit.BLUE),
    GREEN_4(4, Suit.GREEN),      RED_4(4, Suit.RED),      BLACK_4(4, Suit.BLACK),      BLUE_4(4, Suit.BLUE),
    GREEN_5(5, Suit.GREEN),      RED_5(5, Suit.RED),      BLACK_5(5, Suit.BLACK),      BLUE_5(5, Suit.BLUE),
    GREEN_6(6, Suit.GREEN),      RED_6(6, Suit.RED),      BLACK_6(6, Suit.BLACK),      BLUE_6(6, Suit.BLUE),
    GREEN_7(7, Suit.GREEN),      RED_7(7, Suit.RED),      BLACK_7(7, Suit.BLACK),      BLUE_7(7, Suit.BLUE),
    GREEN_8(8, Suit.GREEN),      RED_8(8, Suit.RED),      BLACK_8(8, Suit.BLACK),      BLUE_8(8, Suit.BLUE),
    GREEN_9(9, Suit.GREEN),      RED_9(9, Suit.RED),      BLACK_9(9, Suit.BLACK),      BLUE_9(9, Suit.BLUE),
    GREEN_10(10, Suit.GREEN),    RED_10(10, Suit.RED),    BLACK_10(10, Suit.BLACK),    BLUE_10(10, Suit.BLUE),
    GREEN_JACK(11, Suit.GREEN),  RED_JACK(11, Suit.RED),  BLACK_JACK(11, Suit.BLACK),  BLUE_JACK(11, Suit.BLUE),
    GREEN_QUEEN(12, Suit.GREEN), RED_QUEEN(12, Suit.RED), BLACK_QUEEN(12, Suit.BLACK), BLUE_QUEEN(12, Suit.BLUE),
    GREEN_KING(13, Suit.GREEN),  RED_KING(13, Suit.RED),  BLACK_KING(13, Suit.BLACK),  BLUE_KING(13, Suit.BLUE),
    GREEN_ACE(14, Suit.GREEN),   RED_ACE(14, Suit.RED),   BLACK_ACE(14, Suit.BLACK),   BLUE_ACE(14, Suit.BLUE);

    private static final Set<Card> NORMAL_CARDS = Arrays.stream(values())
            .filter(Card::isNormal)
            .collect(Collectors.toUnmodifiableSet());
    private static final Set<Card> SPECIAL_CARDS = Arrays.stream(values())
            .filter(Card::isSpecial)
            .collect(Collectors.toUnmodifiableSet());
    private static final Map<Integer, Set<Card>> CARDS_BY_VALUE = Arrays.stream(values())
            .filter(Card::isNormal)
            .collect(Collectors.collectingAndThen(
                    Collectors.groupingBy(Card::value, Collectors.toUnmodifiableSet()),
                    Collections::unmodifiableMap
            ));


    private final Integer value;
    private final Suit suit;

    /**
     * Checks whether this card is a special card, i.e. one of {@link #MAHJONG}, {@link #PHOENIX}, {@link #DRAGON} or
     * {@link #HOUND}.
     * @return {@code true} iff this card is a special card.
     * @see #isNormal()
     */
    public boolean isSpecial() {
        return suit == null;
    }

    /**
     * Checks whether this card is a normal card, i.e. none of {@link #MAHJONG}, {@link #PHOENIX}, {@link #DRAGON} or
     * {@link #HOUND}.
     * @return {@code true} iff this card is a normal card.
     * @see #isSpecial()
     */
    public boolean isNormal() {
        return suit != null;
    }

    /**
     * Returns the value of this card. The value is
     * <ul>
     *     <li>{@code 0} for {@link #HOUND}</li>
     *     <li>{@code 1} for {@link #MAHJONG}</li>
     *     <li>{@code null} for {@link #PHOENIX}</li>
     *     <li>{@code 15} for {@link #DRAGON}</li>
     *     <li>obvious for any {@linkplain #isNormal() normal card}</li>
     * </ul>
     * @return the value of this card
     */
    public @Range(from = 1, to = 15) Integer value() {
        return value;
    }

    /**
     * Returns the suit of this {@linkplain #isNormal() normal card} or {@code null} for
     * {@linkplain #isSpecial() special cards}.
     * @return the suit of this card
     */
    public Suit suit() {
        return suit;
    }

    /**
     * Returns an unmodifiable set of all {@linkplain #isNormal() normal cards}.
     * @return a set of cards
     */
    public static @Unmodifiable @NotNull Set<@NotNull Card> getNormalCards() {
        return NORMAL_CARDS;
    }

    /**
     * Returns an unmodifiable set of all {@linkplain #isSpecial() special cards}.
     * @return a set of cards
     */
    public static @Unmodifiable @NotNull Set<@NotNull Card> getSpecialCards() {
        return SPECIAL_CARDS;
    }

    /**
     * Returns an unmodifiable set of all {@linkplain #isNormal() normal cards} with the specified
     * {@linkplain #value() value}.
     * @param value the card value between {@code 2} and {@code 14} (inclusive)
     * @return a set of cards
     */
    public static @Unmodifiable @NotNull Set<@NotNull Card> getCardsByValue(@Range(from = 2, to = 14) int value) {
        return CARDS_BY_VALUE.getOrDefault(value, Collections.emptySet());
    }
}
