package eu.jonahbauer.tichu.common.model;

import eu.jonahbauer.tichu.common.model.combinations.*;
import eu.jonahbauer.tichu.common.model.exceptions.InvalidCombinationException;
import lombok.RequiredArgsConstructor;
import org.jetbrains.annotations.NotNull;

import java.util.*;
import java.util.function.Function;

/**
 * A type of combination.
 */
@RequiredArgsConstructor
public enum CombinationType {
    /**
     * A single card, {@linkplain Card#isNormal() normal} or {@linkplain Card#isSpecial() special}. This is the
     * correct {@code CombinationType} for playing the {@link Card#HOUND} or {@link Card#DRAGON}.
     */
    SINGLE(Single.class, Single::of) {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX, Card.MAHJONG, Card.DRAGON, Card.HOUND);

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 1, SPECIAL_CARDS);
        }
    },
    /**
     * A pair, i.e. two {@linkplain Card#isNormal() normal cards} of the same {@linkplain Card#value() value} or the
     * {@link Card#PHOENIX} and a normal card.
     */
    PAIR(Pair.class, Pair::of) {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);
        private static final int[] INDICES = new int[] {0, 1};

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 2, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES);
        }
    },
    /**
     * A triple, i.e. three {@linkplain Card#isNormal() normal cards} of the same {@linkplain Card#value() value}
     * or the {@linkplain Card#PHOENIX} and two normal cards of the same value.
     */
    TRIPLE(Triple.class, Triple::of) {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);
        private static final int[] INDICES = new int[] {0, 1, 2};

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 3, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES);
        }
    },
    /**
     * A full house, i.e. five cards of which three form a {@link #TRIPLE} and the other two form a {@link #PAIR}.
     * @apiNote A list of cards representing a full house must have the triple at the first three indices and the pair
     * at the last two. This prevents any ambiguity, e.g. when the full house consists of two pairs and the phoenix.
     */
    FULL_HOUSE(FullHouse.class, FullHouse::of) {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);

        private static final int[] INDICES_TRIPLE = new int[] {0, 1, 2};
        private static final int[] INDICES_PAIR = new int[] {3, 4};

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 5, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES_TRIPLE)
                    && isSameValue(cards, INDICES_PAIR);
        }
    },
    /**
     * A sequence, i.e. at least five cards with increasing values. The first card in the sequence may be a
     * {@link Card#MAHJONG} and one of the cards may be the {@link Card#PHOENIX}.
     * @apiNote A list of cards representing a sequence must be sorted by ascending value. This allows to distinguish
     * between a sequence starting with the phoenix and a sequence ending with the phoenix.
     */
    SEQUENCE(SingleSequence.class, SingleSequence::of) {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX, Card.MAHJONG);

        private int getStartValue(@NotNull List<@NotNull Card> cards) {
            return cards.get(0) == Card.PHOENIX
                    ? cards.get(1).value() - 1
                    : cards.get(0).value();
        }

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            if (!checkBasics(cards, null, SPECIAL_CARDS)) return false;
            if (cards.size() < 5) return false;

            int start = getStartValue(cards);

            for (int i = 0; i < cards.size(); i++) {
                var card = cards.get(i);
                if (card == Card.PHOENIX) {
                    if (start + i > 14) {
                        return false;
                    }
                } else if (card.value() != start + i) {
                    return false;
                }
            }

            return true;
        }
    },
    /**
     * A sequence of pairs, i.e. at least two pairs with increasing values.
     * @apiNote Although not strictly necessary, for consistency with {@link #SEQUENCE} a list of cards representing
     * a sequence of pairs must be sorted by ascending value.
     */
    PAIR_SEQUENCE(PairSequence.class, PairSequence::of) {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);

        private int getStartValue(@NotNull List<@NotNull Card> cards) {
            return cards.get(0) == Card.PHOENIX
                    ? cards.get(1).value()
                    : cards.get(0).value();
        }

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            if (!checkBasics(cards, null, SPECIAL_CARDS)) return false;
            if (cards.size() < 4 || cards.size() % 2 == 1) return false;

            int start = getStartValue(cards);

            for (int i = 0; i < cards.size(); i++) {
                var card = cards.get(i);

                if (card != Card.PHOENIX && card.value() != start + i / 2) {
                    return false;
                }
            }

            return true;
        }
    },
    /**
     * A bomb, i.e. four {@linkplain Card#isNormal() normal cards} of the same {@linkplain Card#value() value}.
     */
    BOMB(BombTuple.class, BombTuple::of) {
        private static final int[] INDICES = new int[] {0, 1, 2, 3};
        private static final Set<Card> SPECIAL_CARDS = Collections.emptySet();

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 4, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES);
        }
    },
    /**
     * A bomb sequence, i.e. a {@linkplain Card#suit() monochrome} {@link #SEQUENCE} that does not contain any
     * {@linkplain Card#isSpecial() special cards}.
     * @apiNote Although not strictly necessary, for consistency with {@link #SEQUENCE} a list of cards representing a
     * bomb sequence must be sorted by ascending value.
     * @apiNote
     */
    BOMB_SEQUENCE(BombSequence.class, BombSequence::of) {
        private static final Set<Card> SPECIAL_CARDS = Collections.emptySet();

        private int getStartValue(@NotNull List<@NotNull Card> cards) {
            return cards.get(0).value();
        }

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            if (!checkBasics(cards, null, SPECIAL_CARDS)) return false;
            if (cards.size() < 5) return false;
            if (cards.stream().map(Card::suit).distinct().count() != 1) return false;

            int start = getStartValue(cards);
            for (int i = 1; i < cards.size(); i++) {
                var card = cards.get(i);
                if (card.value() != start + i) {
                    return false;
                }
            }

            return true;
        }
    };

    private final @NotNull Class<? extends Combination> combinationClass;
    private final @NotNull Function<@NotNull List<@NotNull Card>, @NotNull Combination> factory;

    /**
     * Checks whether the given cards form a valid combination of this type. See the documentation of a specific
     * {@code CombinationType} for more information on what cards form a combination of that particular type.
     * @param cards a list of cards
     * @return {@code true} if the cards form a valid combination of this type.
     * @throws NullPointerException if the list or any of its elements are {@code null}.
     */
    public abstract boolean check(@NotNull List<@NotNull Card> cards);

    /**
     * Creates a new combination of this type consisting of the given cards.
     * @param cards a list of cards forming a {@linkplain #check(List) valid} combination
     * @return a new combination.
     * @throws InvalidCombinationException if the given cards do not form a valid combination.
     * @throws NullPointerException if the list or any of its elements are {@code null}.
     */
    public @NotNull Combination of(@NotNull List<@NotNull Card> cards) {
        return factory.apply(cards);
    }

    /**
     * Creates a new combination of this type consisting of the given cards.
     * @param cards a list of cards forming a {@linkplain #check(List) valid} combination
     * @return a new combination.
     * @throws InvalidCombinationException if the given cards do not form a valid combination.
     * @throws NullPointerException if the list or any of its elements are {@code null}.
     */
    public @NotNull Combination of(@NotNull Card @NotNull... cards) {
        return of(List.of(cards));
    }

    /**
     * Returns the class representing a combination of this type.
     * @return the class representing a combination of this type.
     */
    public @NotNull Class<? extends Combination> getCombinationClass() {
        return combinationClass;
    }

    //<editor-fold desc="Utility Methods" defaultstate="collapsed">
    /**
     * Performs basic checks on a list of cards, i.e. this method checks
     * <ul>
     *     <li>that the list does not contain any duplicate cards</li>
     *     <li>when {@code count != null}, that the list is of size {@code count}</li>
     *     <li>
     *         that the list does not contain any {@linkplain Card#isSpecial() special card} other than those in
     *         {@code special}
     *     </li>
     * </ul>
     * @param cards the list of cards
     * @param count the expected number of cards, or {@code null} when the number of cards should not be checked
     * @param special the set of allowed special cards or {@code null}, when all special cards are allowed.
     */
    private static boolean checkBasics(@NotNull List<@NotNull Card> cards, Integer count, @NotNull Set<@NotNull Card> special) {
        if (count != null && cards.size() != count) return false;
        if (containsDuplicates(cards)) return false;

        for (Card card : cards) {
            if (card.isSpecial() && !special.contains(card)) {
                return false;
            }
        }

        return true;
    }

    /**
     * Checks whether the list contains any duplicate cards, i.e. there exists {@code a} and {@code b} such that
     * <pre><code>cards.contains(a) && cards.contains(b) && a.equals(b)</code></pre>
     * @param cards a list of cards
     * @return {@code true} iff the list contains duplicates.
     */
    private static boolean containsDuplicates(@NotNull List<@NotNull Card> cards) {
        if (cards.size() <= 1) return false;
        var set = EnumSet.noneOf(Card.class);
        for (Card card : cards) {
           if (!set.add(card)) return true;
        }
        return false;
    }

    /**
     * Checks that the cards at the specified indices all have the same value.
     * {@linkplain Card#isSpecial() Special cards} are ignored.
     * @param cards a list of cards
     * @param indices an array of indices of cards in the list
     * @return {@code true} iff all of the specified cards have the same value
     */
    private static boolean isSameValue(@NotNull List<@NotNull Card> cards, int @NotNull... indices) {
        Integer value = null;

        for (int index : indices) {
            Card card = cards.get(index);
            if (card.isSpecial()) continue;

            if (value == null) {
                value = card.value();
            } else if (!value.equals(card.value())) {
                return false;
            }
        }

        return true;
    }
    //</editor-fold>
}
