package eu.jonahbauer.tichu.common.model;

import eu.jonahbauer.tichu.common.model.combinations.Combination;
import eu.jonahbauer.tichu.common.model.exceptions.IncompatibleCombinationException;
import eu.jonahbauer.tichu.common.model.exceptions.TooLowCombinationException;
import lombok.EqualsAndHashCode;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Range;
import org.jetbrains.annotations.Unmodifiable;

import java.util.*;

@EqualsAndHashCode
public final class Stack {
    private static final Stack EMPTY = new Stack(Collections.emptyList());

    private final @Unmodifiable @NotNull List<@NotNull Combination> combinations;

    /**
     * Builds a stack of the given combinations.
     * @param combinations an array of combinations from bottom to top
     * @return a stack of the given combinations.
     * @throws IncompatibleCombinationException if any one of the combinations
     *         {@link Combination#isCompatibleWith(Stack) is incompatible with} the stack of all the combinations before it
     * @throws TooLowCombinationException if any one of the combinations
     *         {@linkplain Combination#isHigherThan(Stack) is lower than} the stack of alle the combinations before it
     * @throws NullPointerException if the array or any of its elements are {@code null}
     */
    public static @NotNull Stack of(@NotNull Combination @NotNull... combinations) {
        return of(List.of(combinations));
    }

    /**
     * Builds a stack of the given combinations.
     * @param combinations a list of combinations from bottom to top
     * @return a stack of the given combinations.
     * @throws IncompatibleCombinationException if any one of the combinations
     *         {@link Combination#isCompatibleWith(Stack) is incompatible with} the stack of all the combinations before it
     * @throws TooLowCombinationException if any one of the combinations
     *         {@linkplain Combination#isHigherThan(Stack) is lower than} the stack of alle the combinations before it
     * @throws NullPointerException if the list or any of its elements are {@code null}
     */
    public static @NotNull Stack of(@NotNull List<@NotNull Combination> combinations) {
        Stack out = new Stack(List.copyOf(combinations)); // List#copyOf performs null-checks
        
        for (int i = 0; i < out.size(); i++) {
            Stack lower = out.substack(0, i);
            Combination top = out.get(i);

            if (!top.isCompatibleWith(lower)) {
                throw new IncompatibleCombinationException(lower, top);
            } else if (!top.isHigherThan(lower)) {
                throw new TooLowCombinationException(lower, top);
            }
        }

        return out;
    }

    /**
     * Returns an empty stack.
     * @return an empty stack.
     */
    public static @NotNull Stack empty() {
        return EMPTY;
    }
    
    private Stack(@Unmodifiable @NotNull List<@NotNull Combination> combinations) {
        this.combinations = combinations;
    }

    @Contract(value = "_, _ -> new", pure = true)
    public @NotNull Stack substack(int fromIndex, int toIndex) {
        return new Stack(combinations.subList(fromIndex, toIndex));
    }

    /**
     * Returns the number of elements on this stack.
     * @return the number of elements on this stack
     */
    @Contract(pure = true)
    public int size() {
        return combinations.size();
    }

    /**
     * Returns {@code true} if this stack contains no elements.
     * @return {@code true} if this stack contains no elements
     */
    @Contract(pure = true)
    public boolean isEmpty() {
        return combinations.isEmpty();
    }

    /**
     * Returns the {@code index}-th Combination from the bottom of this stack.
     * @param index the index
     * @return a combination on this stack
     * @throws IndexOutOfBoundsException if the index is out of range ({@code index < 0 || index >= size()})
     */
    @Contract(pure = true)
    public @NotNull Combination get(@Range(from = 0, to = Integer.MAX_VALUE) int index) {
        return combinations.get(index);
    }

    /**
     * Returns the topmost {@link Combination} on this stack. This is equivalent to {@link #top(int) top(0)}.
     * @return the topmost combination
     * @throws IndexOutOfBoundsException if the stack is {@linkplain #isEmpty() empty}
     */
    @Contract(pure = true)
    public @NotNull Combination top() {
        return top(0);
    }

    /**
     * Returns the ({@code -index})-th {@link Combination} from the top of this stack, i.e. {@code top(0)} returns the
     * topmost combination, {@code top(-1)} returns the second-topmost combination, and so on.
     * @param index the index
     * @return a combination on this stack
     * @throws IndexOutOfBoundsException if the index is out of range ({@code index > 0 || index <= -size()})
     */
    @Contract(pure = true)
    public @NotNull Combination top(@Range(from = Integer.MIN_VALUE, to = 0) int index) {
        if (size() <= -index) throw new IndexOutOfBoundsException(index);
        return combinations.get(combinations.size() - 1 + index);
    }

    /**
     * Checks whether the {@code combination} {@linkplain Combination#isCompatibleWith(Stack) is compatible with}
     * this stack.
     * @param combination a combination
     * @return {@code true} iff the {@code combination} is compatible
     * @see Combination#isCompatibleWith(Stack)
     * @see #isHigher(Combination)
     * @see #isCompatibleAndHigher(Combination)
     */
    @Contract(pure = true)
    public boolean isCompatible(@NotNull Combination combination) {
        Objects.requireNonNull(combination, "combination must not be null");
        return combination.isCompatibleWith(this);
    }

    /**
     * Checks whether a {@linkplain #isCompatible(Combination) compatible} {@code combination}
     * {@linkplain Combination#isHigherThan(Stack) is higher than} the {@linkplain #top() topmost} combination
     * on this stack.
     * The behaviour of this method is undefined if the {@code combination} is incompatible and an exception may be thrown.
     * @param combination a combination
     * @return {@code true} iff the combination is higher than the topmost combination on this stack
     * @see Combination#isHigherThan(Stack)
     * @see #isCompatible(Combination)
     * @see #isCompatibleAndHigher(Combination)
     */
    @Contract(pure = true)
    public boolean isHigher(@NotNull Combination combination) {
        Objects.requireNonNull(combination, "combination must not be null");
        return combination.isHigherThan(this);
    }

    /**
     * Checks whether a {@code combination} can be played onto this stack, i.e. the {@code combination} is
     * {@linkplain #isCompatible(Combination) compatible} with this stack and {@linkplain #isHigher(Combination) higher}
     * than the {@linkplain #top() topmost} combination on this stack.
     * @param combination a combination
     * @return {@code true} iff the combination is compatible with and higher than this stack
     * @see Combination#isCompatibleAndHigher(Stack)
     * @see #isHigher(Combination)
     * @see #isCompatible(Combination)
     */
    @Contract(pure = true)
    public boolean isCompatibleAndHigher(@NotNull Combination combination) {
        Objects.requireNonNull(combination, "combination must not be null");
        return combination.isCompatibleAndHigher(this);
    }

    /**
     * Puts the {@code combination} on top of this stack and returns a new, modified stack.
     * @param combination a combination
     * @return a new, modified stack
     * @throws IncompatibleCombinationException if the {@code combination} is
     *         {@linkplain #isCompatible(Combination) incompatible}
     * @throws IllegalArgumentException if the {@code combination} is not {@linkplain #isHigher(Combination) higher}
     *         than the combination currently on top of this tack
     */
    @Contract(value = "_ -> new", pure = true)
    public @NotNull Stack put(@NotNull Combination combination) {
        Objects.requireNonNull(combination, "combination must not be null");
        if (!isCompatible(combination)) {
            throw new IncompatibleCombinationException(this, combination);
        } else if (!isHigher(combination)) {
            throw new TooLowCombinationException(this, combination);
        }

        // copy
        var arr = combinations.toArray(new Combination[combinations.size() + 1]);
        arr[arr.length - 1] = combination;
        return new Stack(Arrays.asList(arr));
    }

    @Override
    public @NotNull String toString() {
        StringBuilder out = new StringBuilder("Stack[");
        for (int i = combinations.size(); i --> 0;) {
            out.append("\n    ").append(combinations.get(i));
        }
        out.append("\n]");
        return out.toString();
    }
}
