package eu.jonahbauer.tichu.common.model.combinations;

import eu.jonahbauer.tichu.common.model.Card;
import eu.jonahbauer.tichu.common.model.CombinationType;
import eu.jonahbauer.tichu.common.model.Stack;
import eu.jonahbauer.tichu.common.model.exceptions.InvalidCombinationException;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Range;
import org.jetbrains.annotations.Unmodifiable;

import java.io.Serializable;
import java.util.List;

public sealed interface Combination extends Serializable permits AbstractCombination, Bomb {
    /**
     * Returns the type of this combination.
     * @return the type of this combination.
     */
    @NotNull CombinationType type();

    /**
     * Returns the cards composing this combination.
     * @return the cards composing this combination.
     */
    @Unmodifiable @NotNull List<@NotNull Card> cards();

    /**
     * Returns the card at the specified index. This is shorthand for {@link #cards()}{@code .}{@link List#get(int) get(index)}.
     * @param index the index of the card to return
     * @return the card at the specified index.
     * @throws IndexOutOfBoundsException if the index is out of bounds ({@code index < 0 || index >= size()})
     */
    @NotNull Card card(@Range(from = 0, to = 13) int index);

    /**
     * Returns the size of this combination, i.e. the number of cards it consists of. This is shorthand for
     * {@link #cards()}{@code .}{@link List#size() size()}.
     * @return the size of this combination.
     */
    @Range(from = 1, to = 14) int size();

    /**
     * Checks whether this combination is <em>compatible</em> with the given stack. The definition of
     * <em>compatible</em> depends on the type of this combination:
     * <ul>
     *     <li>
     *         a {@link Bomb} is compatible with any stack that does not have a {@linkplain Single single}
     *         {@link Card#HOUND} on top.
     *     </li>
     *     <li>
     *         any other combination is compatible with an empty stack and all stacks that have a combination
     *         of the same {@link #type()} and {@link #size()} on top.
     *     </li>
     * </ul>
     * @param stack a stack of combinations
     * @return {@code true}, iff this combination is compatible with the given stack.
     */
    boolean isCompatibleWith(@NotNull Stack stack);

    /**
     * Checks whether this combination is <em>higher</em> than the given stack, i.e. it could be played on top of
     * that stack. This method does not, however, ensure that this combination <em>can</em> be played on top of that
     * stack in every circumstance, since mahjong-wishes and right to play are not accounted for.
     *
     * <p>This method assumes that this combination {@linkplain #isCompatibleWith(Stack) is compatible with} the
     * given stack, otherwise the result is undefined and the method may throw an exception.
     * @param stack a stack of combinations
     * @return {@code true}, iff this combination is higher than the given stack.
     */
    boolean isHigherThan(@NotNull Stack stack);

    /**
     * Checks whether this combination is {@linkplain #isCompatibleWith(Stack) compatible with} and
     * {@linkplain #isHigherThan(Stack) higher than} the given stack.
     * @param stack a stack of combinations
     * @return {@code true}, iff this combination is compatible with and higher than the given stack.
     */
    boolean isCompatibleAndHigher(@NotNull Stack stack);

    /**
     * Creates a new combination of the given type consisting of the given cards.
     * @param cards a list of cards forming a {@linkplain CombinationType#check(List) valid} combination
     * @param type a combination type
     * @return a new combination.
     * @throws InvalidCombinationException if the given cards do not form a valid combination.
     * @throws NullPointerException if the type, the list or any of its elements are {@code null}.
     */
    @Contract("_, _ -> new")
    static Combination of(@NotNull CombinationType type, @NotNull List<@NotNull Card> cards) {
        return switch (type) {
            case SINGLE -> Single.of(cards);
            case PAIR -> Pair.of(cards);
            case TRIPLE -> Triple.of(cards);
            case FULL_HOUSE -> FullHouse.of(cards);
            case SEQUENCE -> SingleSequence.of(cards);
            case PAIR_SEQUENCE -> PairSequence.of(cards);
            case BOMB -> BombTuple.of(cards);
            case BOMB_SEQUENCE -> BombSequence.of(cards);
        };
    }
}
